--[[========================================================================================
	
	
	Author: Ferodra [Arenima - Alleria EU]
		Email: ferodra@gmx.de

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
    ========================================================================================]]

--[[
	This Addon provides a big, dynamic library full of methods to instantly create unitframes for every need.

	"local *" and "E.*" explaination:
		"local" defines the private scope in LUA
		"E" is our 'class' name in this case and lets us access everything defined within.
	
	Since we want to access some variables across our AddOn, we have to throw them into this private(/public) scope (local(/E)).
	
	Important Lua-Garbage note:
		Setting the value of any table via {}, creates a NEW table and will contribute to generating garbage!!!
		To properly do this, empty a table with "wipe(t)" and settings values with: "table.val = newvalue"
		Not table = {val = newvalue}
]]-- 

-------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------  Globals  -------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------

local E = select(2, ...) -- Engine
local CO,LOC,UF,AB,LT,AUR,TT,BC,B = E:LoadModules("Config", "Locale", "Unitframes", "Actionbars", "Layout", "Auras", "Tooltip", "Bar_Cast", "Blizzard")

E.E 									= 			CreateFrame("Frame") -- EngineEvent
E.InitComplete							=			false
E.Media									=			{}

E.Movers = {}
E.CVars = {}

E.UNIT_MAXLEVEL = GetMaxPlayerLevel()

E.EventUnits = {
	["PLAYER_TARGET_CHANGED"]	=	"target",
	["PLAYER_FOCUS_CHANGED"]	=	"focus",
}

-------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------

local _
local C	= CreateFrame("Frame")

local UnitIsPlayer 	= UnitIsPlayer
local UnitClass 	= UnitClass
local UnitIsFriend 	= UnitIsFriend
local UnitIsEnemy 	= UnitIsEnemy

-----------------------------------------------------------------------------------------
-- Fix for BfA changes. We have to access some more functions via the namespace now!
-----------------------------------------------------------------------------------------
	E.GetPlayerMapPosition		= C_Map.GetPlayerMapPosition								----- Returns the units (arg2) X and Y position for mapID (arg1)
	E.GetCurrentMapID			= C_Map.GetBestMapForUnit									----- Returns the player map ID when "player" is passed
-----------------------------------------------------------------------------------------

local EventUnit
local HighPowers = {0} -- Powers that should not be displayed separately

-- Prints AddOn messages to console/chat
function E:print(msg)
	print(E.PrintPrefix .. E.MessageColor .. " " .. msg .. E.ColorReset)
end

-- Prints AddOn debug-messages to console/chat. But without any concatentation to prevent errors caused by nil
-- Messages can be turned off by simply setting E.Debug to false
function E:debugprint(...)
	if E.Debug == true then
		print(E.PrintPrefix .. E.ColorReset, ...)
	end
end

-- Check if power shouldn't be displayed separated
function E:IsHighPower(powerId)
	for k,v in pairs(HighPowers) do
		if v == powerId then return true end
	end
	
	return false
end

-- Retrieve class color of unit (@param1)
function E:GetUnitClassColor(unit)
	return E.ClassColors[select(3,UnitClass(unit))]
end

function E:GetUnitPowerColor(unit)
	return CO.db.profile.colors.powers[UnitPowerType(unit)] or E.PowerColors[UnitPowerType(unit)]
end

local UnitReactionDefault, UnitReactionColor, UnitReactionClassID, UnitReactionProfileTarget, UnitReactionReaction, ReturnTable
ReturnTable = {}
UnitReactionDefault = {1, 1, 1}

function E:GetUnitReactionColor(Unit)
	UnitReactionColor = UnitReactionDefault
	_, _, UnitReactionClassID = UnitClass(Unit)
	UnitReactionProfileTarget = CO.db.profile.unitframe.units.all.reactionColors
		
	UnitReactionReaction 		= UnitReaction(Unit, "player") -- Get reaction to player
	if UnitReactionReaction then
		if not UnitIsPlayer(Unit) then
			
			if not UnitIsFriend(Unit, "player") and UnitIsEnemy(Unit, "player") then
				UnitReactionReaction = 1
			end
			
			if UnitReactionReaction >= 5 then
				UnitReactionColor = UnitReactionProfileTarget["friendly"]
			elseif UnitReactionReaction == 4 then
				UnitReactionColor = UnitReactionProfileTarget["neutral"]
			elseif UnitReactionReaction == 3 then
				UnitReactionColor = UnitReactionProfileTarget["unfriendly"]
			else
				UnitReactionColor = UnitReactionProfileTarget["hostile"]
			end
		else
			if UnitReactionReaction >= 5 then  
				UnitReactionColor = E.ClassColors[UnitReactionClassID]
			else
				UnitReactionColor = UnitReactionProfileTarget["hostile"]
			end
		end
	elseif UnitReactionClassID then
		UnitReactionColor = E.ClassColors[UnitReactionClassID]
	end
	
	ReturnTable.r = UnitReactionColor[1]
	ReturnTable.g = UnitReactionColor[2]
	ReturnTable.b = UnitReactionColor[3]
	
	
	return ReturnTable
end

local BorderParent
function E:SkinButtonIcon(B, BC)
	if not B.Border then
		BorderParent = B:GetParent()
		
		B.Border = CreateFrame("Frame", nil)
		B.Border:SetAllPoints(BorderParent)
		B.Border:SetParent(BorderParent)
		
		B.Border:SetBackdrop({bgFile = "", 
			edgeFile = [[Interface\Buttons\WHITE8X8]], 
			edgeSize = 1, 
			tile = true, tileSize = 16});
			
		B.HighlightOverlay = BorderParent:CreateTexture(nil, "HIGHLIGHT")
		B.HighlightOverlay:SetAllPoints(BorderParent)
		
		B.HighlightOverlay:SetColorTexture(1,1,0, 0.15)
		B.HighlightOverlay:SetBlendMode("ADD")
		
		B:SetTexCoord(0.06,0.94,0.06,0.94)
	end
	if BC then
		if not BC.r then
			B.Border:SetBackdropBorderColor(BC[1], BC[2], BC[3], 1)
		else
			B.Border:SetBackdropBorderColor(BC.r, BC.g, BC.b, 1)
		end
	end
end

function E:ToggleParty(State)
	E:ToggleFrame(E.Frames["PartyMainFrame"], State, false)
end

function E:SetVisibilityHandler(object)
	object:SetAttribute("_onstate-visible", [[
		if newstate == 1 then
			self:Show();
			self:SetAttribute("IsShown", true)
		else
			self:Hide();
			self:SetAttribute("IsShown", false)
		end
		-- print(newstate, self:GetName())
	]]);
end

function E:ResetMoverPositions()
	CO.db.profile.movers = E.MoverDefaults
	
	E:LoadMoverPositions()
end

-- Toggle mover overlays and drag functionality
-- We can NOT simply show the whole thing, since we still have to hide them afterwards
-- this results in EVERY frame disappearing
function E:ToggleMover(state)	
		
	-- Use Frame:IsEventRegistered("event") to check OnEvent visiblity for special frames
	
	--local RRD = E:GetModule("RaidRoleData")
	
	-- Here we toggle special blizzard frames
	if state == true then
		-- Extra Button
		if not HasExtraActionBar() then
			ExtraActionBarFrame:Show()
			ExtraActionBarFrame:SetAlpha(1) -- W.h.y
			ExtraActionButton1:Show()
		end
		-- 
		--BNToastFrame:Show()
		VehicleSeatIndicator:Show()
		--if RRD:GetAttribute("IsShown") == false then RRD:Show() end
	else
		-- Extra Button
		-- Prevent the button from being hidden when it actually is supposed to stay active
		if not HasExtraActionBar() then
			ExtraActionBarFrame:Hide()
			ExtraActionBarFrame:SetAlpha(0) -- W.h.y
			ExtraActionButton1:Hide()
		end
		--
		--BNToastFrame:Hide()
		VehicleSeatIndicator:Hide()
		--if RRD:GetAttribute("IsShown") == false then RRD:Hide() end
	end
	B:ToggleZoneAbility(state)
	B:ToggleAltPowerBar(state)
	--UF:OverrideHolderVisibility(state)
	
	-- Castbar movers are a bit special
	--BC:ToggleMovers(state)
	
	-- k: Mover Name, v: Mover Object
	for k,v in pairs(E.Movers) do	
		if state == true and v.Frame.MoverEnabled then
			v.Handle:EnableMouse(true)
			v.Handle:Show()
		else
			v.Handle:EnableMouse(false)
			v.Handle:Hide()
		end
	end
end

-- Returns a registered mover from child object or name-string
function E:GetMover(C)
	if type(C) == "string" then
		return E.Movers[C .. "Mover"]
	else
		return E.Movers[E:GetFullFrameName(C) .. "Mover"]
	end
end

function E:RegisterMover(M, MName)
	E.Movers[MName] = M
end

function E:LoadMoverPositions(limit)
	local Conf = CO.db.profile.movers
	local ConfData = {}
	
	if limit then
		local mover = E:GetMover(limit)
		
		if mover then
			ConfData = Conf[mover:GetName()]
			E:RepositionMover(mover, ConfData["point"], ConfData["relativePoint"], ConfData["xOffset"], ConfData["yOffset"])
			
			return
		end
	end
	
	-- k: Mover Name - v: Mover Object
	for k,v in pairs(E.Movers) do
		ConfData = Conf[k]
		
		if ConfData then
			if v then
				E:RepositionMover(v, ConfData["point"], ConfData["relativePoint"], ConfData["xOffset"], ConfData["yOffset"])
			else
				E:print("WARNING: Corrupt mover data found!")
			end
		end
	end
end

-- @PARAM1: Child target
-- @PARAM2: Localized mover name for user display
function E:CreateMover(C, LT, A, X, Y)
	local MNameRaw = E:GetFullFrameName(C)
	local MName = MNameRaw .. "Mover"
	
	if not A then A = "CENTER" end
	if not (X and Y) then X, Y = C:GetWidth(), C:GetHeight() if X <= 0 and Y <= 0 then X, Y = 50, 50 end end
	local M = E:NewFrame("Frame", MName, "LOW", X,Y, {"CENTER", E.Parent, "CENTER", 0, 0}, E.Parent)	
	
	-- Position update function
	-- This is probably the only existing way to deal with constantly moving blizzard frames and also work for everything else ofc
	local function Mover_SetPosition(_, _, parent)
		if parent ~= M then
			C:ClearAllPoints()
			C:SetParent(M)
			C:SetPoint(A, M, A, 0, 0)
		end
	end
	hooksecurefunc(C, "SetPoint", Mover_SetPosition)
	
	C:SetPoint("CENTER") -- Execute hook initially to make it work from the beginning [was eating some time to figure that out]
	
	-- Create Mover handle to interact with
	M.Handle = E:CreateMoverHandle(C, LT, A, X, Y)
	M.Handle:SetParent(M)
	M.Handle:SetFrameLevel(9999)
	M.Handle:SetFrameStrata("TOOLTIP")
	
	-- Store reference to child frame
	C.MoverEnabled = true
	M.Frame = C
	
	-- E:GetMover(FrameObjectOrNameAsString).Handle:Show() -- We can use this to access the handle at any time!
	
	M.Handle:SetScript("OnDragStart", function(self)
		local parent = M
		parent:SetMovable(true)
		self:SetClampedToScreen(false)
		parent:SetClampedToScreen(false)
		parent:StartMoving()
	end)
	M.Handle:SetScript("OnDragStop", function(self)
		local point, relativePoint, xOfs, yOfs
		local parent = M
		local title = MName
		
		parent:SetMovable(false)
		parent:StopMovingOrSizing()
		
		-- Fix for movers without any default values
		if not CO.db.profile.movers[title] then
			CO.db.profile.movers[title] = {}
		end
		local conf = CO.db.profile.movers[title]
		
		
		point, _, relativePoint, xOfs, yOfs = parent:GetPoint(parent:GetNumPoints())
		conf["point"] 			= point
		conf["relativePoint"] 	= relativePoint
		conf["xOffset"] 		= xOfs
		conf["yOffset"] 		= yOfs
	end)
	
	-- Internal register
	-- MoverObject, MoverName
	E:RegisterMover(M, MName)
	
	return M
end

function E:RepositionMover(M, Point, RelativePoint, OffsetX, OffsetY)
	E:RepositionFrame(M, Point, RelativePoint, OffsetX, OffsetY, E.Parent)
end

-- Since we move frames via movers, which are basically frames, the frames we want to move with are parented to the mover
-- By later "copying" the translations made to the overlay to the base parent, we can move the whole thing
function E:CreateMoverHandle(C, LT, A, X, Y)
	local MH = E:NewFrame("Frame", "Handle", "HIGH", X,Y, {A, C, A, 0, 0}, C)
	local RGB = E:GetUnitClassColor("player")
	MH:EnableMouse(true)
	MH:SetMovable(true)
	MH:RegisterForDrag("LeftButton")
	
	MH:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
		edgeFile = [[Interface\Buttons\WHITE8X8]], 
		edgeSize = 1, 
		tile = true, tileSize = 16});
	MH:SetBackdropColor(RGB[1], RGB[2], RGB[3], 1)
	MH:SetBackdropBorderColor(0.15, 0.15, 0.15, 1)
	
	-- Init font overlay
	MH.Name = MH:CreateFontString(nil, "ARTWORK")
	E:InitializeFontFrame(MH.Name, "ARTWORK", "FRIZQT__.TTF", 14, {1,1,1}, 1, {0,0}, "", 250, Y, MH, "CENTER", {1,1})
	MH.Name:ClearAllPoints()
	MH.Name:SetPoint("CENTER", MH, 'CENTER', 5, -5)
	
	MH.Name:SetText(LT) -- Set provided name
	
	MH:Hide()
	
	return MH
end

-- Update Mover dimensions based on parent
function E:UpdateMoverDimensions(C)
	local M = E:GetMover(C)
	
	if not M then return end
	
	M:SetSize(C:GetWidth(), C:GetHeight())
	M.Handle:SetSize(C:GetWidth(), C:GetHeight())
end

E.StatusBars = {}
function E:RegisterStatusBar(Bar)
	-- Expects Bar.Unit and Bar.Type to already be present
	table.insert(E.StatusBars, Bar)	
end

-- Method to initially set all required CVars on an empty/default profile
function E:InitCVars()
	for k,v in pairs(CO.db.profile.CVars) do
		E:RegisterCVar(k,v)
	end
	
	E:LoadRegisteredCVars()
end

function E:RegisterCVar(CVar, value)
	E.CVars[CVar] = value
end

function E:LoadRegisteredCVars()
	for k,v in pairs(E.CVars) do
		SetCVar(k, v)
	end
end

function C:Init()
	
	CO, TT, UF, AUR, BC, B = E:LoadModules("Config", "Tooltip", "Unitframes", "Auras", "Bar_Cast", "Blizzard")
	E:InitCVars()
	
	local arg1
	
	E.E:SetScript("OnEvent",function(self,event, ...)

		-- BfA Version
		--if event == "UNIT_POWER_UPDATE" then E:UpdateUnitInfo("power", arg1); UF:UpdateAlternatePower() end
		-- Legion Version
		if select(3, UnitClass("player")) ~= 6 then
			if event == "UNIT_POWER_UPDATE" or event == "RUNE_POWER_UPDATE" then UF:UpdateAlternatePower() end
		end

		if event == "UNIT_DISPLAYPOWER" or event == "PLAYER_SPECIALIZATION_CHANGED" or event == "SPELLS_CHANGED" then UF:SetupAlternatePower(); end
	end)
end

E:AddModule("Core", C)